from abc import ABC, abstractmethod
import torch
from reetoolbox.constraints import Constraints


def untargeted_loss(outputs, labels):
    loss = outputs.gather(1, labels.unsqueeze(1))[:, 0]
    return loss


def targeted_loss(outputs, targets):
    num_classes = list(outputs.shape)[1]
    all_out = torch.sum(outputs, dim=1)
    target_out = outputs.gather(1, targets)[:, 0]
    loss = (all_out / num_classes) - target_out
    # loss = -target_out
    return loss


class Optimiser(ABC):
    def __init__(self, model, Transform, hyperparameters, transform_hyperparameters, criterion=untargeted_loss,
                 device="cuda:0"):
        self.model = model
        self.Transform = Transform
        self.hyperparameters = hyperparameters
        self.transform_hyperparameters = transform_hyperparameters
        self.device = device
        self.transform = None
        self.criterion = criterion

    @abstractmethod
    def optimise(self):
        pass


class PGD(Optimiser):
    def optimise(self, inputs, targets=None, reset_weights=True):
        epsilon = self.hyperparameters["epsilon"]
        steps = self.hyperparameters["steps"]
        constraint = self.hyperparameters["constraint"]
        C = self.hyperparameters["C"]
        input_range = self.hyperparameters["input_range"]

        if constraint is not None:
            constraint_func = getattr(Constraints, constraint)
            constraints = Constraints

        inputs = inputs.to(self.device)

        if self.transform is None or reset_weights:
            self.transform = self.Transform(input_shape=inputs.shape, device=self.device,
                                            **self.transform_hyperparameters)

        in_train_mode = self.model.training
        self.model.eval()

        grads = []
        for param in self.model.parameters():
            grads.append(param.requires_grad)
            param.requires_grad = False

        opt = torch.optim.RMSprop([self.transform.weights], lr=epsilon)

        for i in range(steps):
            opt.zero_grad()

            if targets is None:
                outputs = self.model(inputs)
                targets = torch.argmax(outputs, dim=1)

            original_inputs = inputs.clone()
            adv_inputs = self.transform.forward(inputs)
            inputs = original_inputs
            adv_outputs = self.model(adv_inputs)

            loss = self.criterion(adv_outputs, targets)

            loss.backward(torch.ones_like(loss), retain_graph=True)
            opt.step()

            if constraint is not None:
                self.transform.weights = constraint_func(constraints, self.transform.weights,
                                                         self.transform.base_weights, C)

        original_inputs = inputs.clone()
        adv_inputs = self.transform.forward(inputs)
        inputs = original_inputs

        for i, param in enumerate(self.model.parameters()):
            param.requires_grad = grads[i]

        if in_train_mode:
            self.model.train()

        return inputs, adv_inputs


class StochasticSearch(Optimiser):
    def optimise(self, inputs, targets=None, reset_weights=True):
        samples = self.hyperparameters["samples"]
        weight_ranges = self.hyperparameters["weight_ranges"]
        input_range = self.hyperparameters["input_range"]

        inputs = inputs.to(self.device)

        if self.transform is None or reset_weights:
            self.transform = self.Transform(input_shape=inputs.shape, device=self.device,
                                            **self.transform_hyperparameters)
            self.best_loss = None
            self.best_adv_inputs = inputs

        in_train_mode = self.model.training
        self.model.eval()

        grads = []
        for param in self.model.parameters():
            grads.append(param.requires_grad)
            param.requires_grad = False

        with torch.no_grad():
            for i in range(samples):
                for j, weight_name in enumerate(weight_ranges):
                    self.transform.weights[weight_name] = torch.FloatTensor(
                        *self.transform.weights[weight_name].shape).uniform_(*weight_ranges[weight_name]).to(self.device)

                if targets is None:
                    outputs = self.model(inputs)
                    targets = torch.argmax(outputs, dim=1)

                original_inputs = inputs.clone()
                adv_inputs = self.transform.forward(inputs)
                inputs = original_inputs
                adv_outputs = self.model(adv_inputs)

                loss = self.criterion(adv_outputs, targets)

                if self.best_loss is None or self.best_adv_inputs is None:
                    self.best_loss = loss.clone()
                    self.best_adv_inputs = adv_inputs.clone()
                else:
                    for j, input_loss in enumerate(loss):
                        if input_loss < self.best_loss[j]:
                            self.best_adv_inputs[j] = adv_inputs[j]
                            self.best_loss[j] = input_loss

        original_inputs = inputs.clone()
        inputs = original_inputs

        for i, param in enumerate(self.model.parameters()):
            param.requires_grad = grads[i]

        if in_train_mode:
            self.model.train()

        return inputs, self.best_adv_inputs



import torch
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import copy
from reetoolbox.optimisers import targeted_loss, untargeted_loss


class Evaluator:
    def __init__(self, model, dataset, dataloader, TransformOptimiser, Transform, optimiser_params, trans_params, criterion=untargeted_loss, device="cuda:0"):
        self.model = model
        self.dataset = dataset
        self.dataloader = dataloader
        self.device = device
        self.trans_params = trans_params
        self.attack = TransformOptimiser(self.model, Transform, optimiser_params, trans_params, criterion=criterion, device=device)
        self.model.eval()
        self.criterion = criterion

    def predict(self, adversarial, perturbation_measure=None, weight_measure=None):
        outputs = []
        adv_outputs = []
        all_labels = []
        pert_measures = []
        weight_measures = []

        count = 0

        for inputs, labels in self.dataloader:
            count += len(inputs)
            inputs, labels = inputs.to(self.device), labels.to(self.device)

            if adversarial:
                inputs, adv_inputs = self.attack.optimise(inputs, targets=labels, reset_weights=True)
                batch_output = self.model(adv_inputs).cpu().detach()
                adv_outputs.extend(batch_output)

                if perturbation_measure is not None:
                    input_pert = (adv_inputs - inputs).cpu().detach()
                    pert_measures.append(perturbation_measure(input_pert))
                if weight_measure is not None:
                    weight_pert = (self.attack.transform.weights - self.attack.transform.base_weights).cpu().detach()
                    weight_measures.append(weight_measure(weight_pert))

            batch_output = self.model(inputs).cpu().detach()
            outputs.extend(batch_output)
            all_labels.extend(labels.cpu().detach())

        return_dict = {}
        return_dict["outputs"] = torch.stack(outputs)
        return_dict["labels"] = torch.stack(all_labels)
        if adversarial:
          return_dict["adversarial_outputs"] = torch.stack(adv_outputs)
        if pert_measures:
            return_dict["perturbation_measures"] = torch.stack(pert_measures)
        if weight_measures:
            return_dict["weight_measures"] = torch.stack(weight_measures)
        return return_dict

    def compute_metric(self, metric, **parameters):
        results = self.predict(**parameters)
        score = metric(results)
        return score

    def metric_vs_strength(self, param, param_range, step_size, metric, **metric_params):
        start = param_range[0]
        end = param_range[1] + step_size
        param_values = [f for f in np.arange(start, end, step_size)]

        all_scores = []

        printerval = np.round(len(param_values) / 4)

        original_parameters = copy.deepcopy(self.attack.hyperparameters)

        for i, value in enumerate(param_values):
            self.attack.hyperparameters[param] = value

            if i == 0:
                print("Starting hyperparameters:", self.attack.hyperparameters)
            if i % printerval == 0:
                print(f"{round(i * 100 / len(param_values))}% complete...")

            score = self.compute_metric(metric, **metric_params)
            all_scores.append(score)

        print("Done.")

        self.attack.hyperparameters = original_parameters

        return param_values, all_scores

    def attack_inputs(self, input_indices, target_classes=None, criterion=None):
        inputs = []
        labels = []

        for i in input_indices:
            inputs.append(self.dataset[i][0])
            labels.append(self.dataset[i][1])

        inputs = torch.stack(inputs).to(self.device)
        labels = torch.stack(labels).to(self.device)

        if target_classes is not None and criterion is not None:
            self.attack.criterion = criterion
            inputs, adv_inputs = self.attack.optimise(inputs, targets=target_classes, reset_weights=True)
            self.attack.criterion = self.criterion
        else:
            inputs, adv_inputs = self.attack.optimise(inputs, targets=labels, reset_weights=True)
        return inputs, adv_inputs

    def set_attack_hyperparameters(self, hyperparameters):
        self.attack.hyperparameters = hyperparameters



import torch.nn.functional as fcn
from abc import ABC, abstractmethod
import torch
from torch.autograd import Variable
import torchvision.transforms.functional as TF
from reetoolbox.utils import *


class Transform(ABC):
    def __init__(self, input_shape, device):
        """
        Abstract Transform class
        :param input_shape: shape of input batch
        :param device: device that is performing the computations, e.g., cpu, cuda:0
        """
        self.input_shape = input_shape
        self.device = device
        self.weights = None

    @abstractmethod
    def forward(self, x):
        """
        Performs the transform on x using self.weights
        :return: x
        """
        pass


class PixelTransform(Transform):
    def __init__(self, input_shape, device, input_range=(0, 255), noise_range=(-0.1, 0.1)):
        super().__init__(input_shape, device)
        self.input_range = input_range
        self.base_weights = torch.zeros(input_shape).to(device)
        weights = torch.FloatTensor(*input_shape).uniform_(*noise_range).to(device)
        self.weights = Variable(weights, requires_grad=True)

    def forward(self, x):
        x = torch.clamp(x + self.weights, *self.input_range)
        return x


class StainTransform(Transform):
    def __init__(self, input_shape, device, input_range=(0, 255), noise_range=(0, 0)):
        super().__init__(input_shape, device)
        self.max_input_value = input_range[1]

        batch_size = self.input_shape[0]
        self.ruiford = np.array([[0.65, 0.70, 0.29],
                                 [0.07, 0.99, 0.11],
                                 [0.27, 0.57, 0.78]], dtype=np.float)
        self.base_weights = np.stack([self.ruiford] * batch_size)
        self.base_weights = torch.tensor(self.base_weights).to(device)

        self.weights = np.stack([self.ruiford] * batch_size)
        self.weights = torch.tensor(self.weights).to(device)
        self.weights += torch.FloatTensor(*self.weights.shape).uniform_(*noise_range).to(device)
        self.weights = Variable(self.weights, requires_grad=True)

    def forward(self, x):
        x = self.run_unmix(x)
        x = self.run_mix(x)
        return x

    def run_unmix(self, x):
        # operates in NCHW
        x = (x / self.max_input_value).float()
        outputs = []
        for i, X in enumerate(x):
            Z = self._stain_unmix(X.permute(1, 2, 0), i)
            outputs.append(Z.permute(2, 0, 1))
        x = torch.stack(outputs)
        x = (x * self.max_input_value).float()
        return x

    def run_mix(self, x):
        x = (x / self.max_input_value).float()
        outputs = []
        for i, X in enumerate(x):
            Z = self._stain_remix(X.permute(1, 2, 0), i)
            outputs.append(Z.permute(2, 0, 1))
        x = torch.stack(outputs)
        x = (x * self.max_input_value).float()
        return x

    def _stain_unmix(self, X, index):
        # operates in HWC
        X = torch.max(X, 1e-6 * torch.ones_like(X))
        Z = (torch.log(X) / np.log(1e-6)) @ torch.inverse(self.normalise_matrix(self.base_weights[index]))
        return Z

    def _stain_remix(self, Z, index):
        # operates in HWC
        log_adjust = -np.log(1E-6)
        log_rgb = -(Z * log_adjust) @ self.normalise_matrix(self.weights[index])
        rgb = torch.exp(log_rgb)
        rgb = torch.clamp(rgb, 0, 1)
        return rgb

    def normalise_matrix(self, matrix):
        return fcn.normalize(matrix).float()

class MeanTransform(Transform):
    def __init__(self, input_shape, device, input_range=(0, 255), noise_range=(-0.1, 0.1)):
        super().__init__(input_shape, device)
        self.input_range = input_range
        batch_size = self.input_shape[0]
        shape = (batch_size, 1)
        self.base_weights = torch.zeros(shape).to(device)
        weights = torch.FloatTensor(*shape).uniform_(*noise_range).to(device)
        self.weights = Variable(weights, requires_grad=True).to(device)

    def forward(self, x):
        for i in range(self.input_shape[0]):
            x[i] = torch.clamp(x[i] + self.weights[i], *self.input_range)
        return x


class CutoutTransform(Transform):
    def __init__(self, input_shape, device, height, width):
        super().__init__(input_shape, device)
        self.height = height
        self.width = width
        batch_size = input_shape[0]
        self.weights = {
            "corner_x": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "corner_y": torch.stack([torch.tensor([0.0])] * batch_size).to(device)
        }

    def forward(self, x):
        for i, _ in enumerate(x):
            corner_y = self.weights["corner_y"][i, 0].int()
            corner_x = self.weights["corner_x"][i, 0].int()
            x[i, :, corner_y:corner_y + self.height, corner_x:corner_x + self.width] = 0
        return x


class RotateTransform(Transform):
    def __init__(self, input_shape, device):
        super().__init__(input_shape, device)
        batch_size = input_shape[0]
        self.weights = {
            "angle": torch.stack([torch.tensor([0.0])] * batch_size).to(device)
        }

    def forward(self, x):
        for i, _ in enumerate(x):
            x[i] = TF.rotate(x[i], self.weights["angle"][i].item())
        return x


class CropTransform(Transform):
    def __init__(self, input_shape, device):
        super().__init__(input_shape, device)
        batch_size = input_shape[0]
        self.weights = {
            "top": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "left": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "height": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "width": torch.stack([torch.tensor([0.0])] * batch_size).to(device)
        }

    def forward(self, x):
        for i, _ in enumerate(x):
            top = int(self.weights["top"][i].item())
            left = int(self.weights["left"][i].item())
            height = int(self.weights["height"][i].item())
            width = int(self.weights["width"][i].item())
            cropped = TF.crop(x[i], top=top, left=left, height=height, width=width)

            left_pad = round((self.input_shape[2] - width) / 2)
            right_pad = self.input_shape[2] - width - left_pad
            top_pad = round((self.input_shape[3] - height) / 2)
            bottom_pad = self.input_shape[3] - height - top_pad
            x[i] = TF.pad(cropped, (left_pad, top_pad, right_pad, bottom_pad))
        return x


class BlurTransform(Transform):
    def __init__(self, input_shape, device):
        super().__init__(input_shape, device)
        batch_size = input_shape[0]
        self.weights = {
            "corner_x": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "corner_y": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "height": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "width": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "kernel_size": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
            "sigma": torch.stack([torch.tensor([0.0])] * batch_size).to(device)
        }

    def forward(self, x):
        for i, _ in enumerate(x):
            corner_y = self.weights["corner_y"][i, 0].int()
            corner_x = self.weights["corner_x"][i, 0].int()
            height = self.weights["height"][i, 0].int()
            width = self.weights["width"][i, 0].int()
            kernel_size = int(self.weights["kernel_size"][i, 0].item())
            if kernel_size % 2 == 0:
                kernel_size += 1
            sigma = self.weights["sigma"][i, 0].item()
            blur_section = TF.gaussian_blur(x[i, :, corner_y:corner_y + height, corner_x:corner_x + width], kernel_size,
                                            sigma)
            x[i, :, corner_y:corner_y + height, corner_x:corner_x + width] = blur_section
        return x


class ZoomInTransform(Transform):
    def __init__(self, input_shape, device):
        super().__init__(input_shape, device)
        batch_size = input_shape[0]
        self.weights = {
            "scale": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
        }

    def forward(self, x):
        for i, _ in enumerate(x):
            scale = self.weights["scale"][i].item()
            big = TF.resize(x[i], (int(self.input_shape[2] * scale), int(self.input_shape[3] * scale)))

            top = int(torch.FloatTensor(1).uniform_(0, big.shape[1] - self.input_shape[2]).to(device).item())
            left = int(torch.FloatTensor(1).uniform_(0, big.shape[2] - self.input_shape[3]).to(device).item())

            x[i] = TF.crop(big, top=top, left=left, height=self.input_shape[2], width=self.input_shape[3])
        return x


class ZoomOutTransform(Transform):
    def __init__(self, input_shape, device):
        super().__init__(input_shape, device)
        batch_size = input_shape[0]
        self.weights = {
            "scale": torch.stack([torch.tensor([0.0])] * batch_size).to(device),
        }

    def forward(self, x):
        for i, _ in enumerate(x):
            scale = self.weights["scale"][i].item()
            small = TF.resize(x[i], (int(self.input_shape[2] * scale), int(self.input_shape[3] * scale)))

            width = small.shape[1]
            height = small.shape[2]
            left_pad = round((self.input_shape[2] - width) / 2)
            right_pad = self.input_shape[2] - width - left_pad
            top_pad = round((self.input_shape[3] - height) / 2)
            bottom_pad = self.input_shape[3] - height - top_pad
            x[i] = TF.pad(small, (left_pad, top_pad, right_pad, bottom_pad))
        return x


class HEDTransform(Transform):
    def __init__(self, input_shape, device, input_range=(0, 255)):
        super().__init__(input_shape, device)
        batch_size = self.input_shape[0]
        self.input_range = input_range
        self.weights = {
            "alpha": torch.zeros(batch_size, 3).to(device),
            "beta": torch.zeros(batch_size, 3).to(device)
        }

        self.max_input_value = input_range[1]
        self.ruifrock = torch.tensor(np.array([[0.65, 0.70, 0.29],
                                 [0.07, 0.99, 0.11],
                                 [0.27, 0.57, 0.78]], dtype=np.float)).to(device)

    def forward(self, x):
        x = (x / self.max_input_value).float()
        x = self.run_unmix(x)
        for i, hed in enumerate(x):
            for j, channel in enumerate(hed):
                hed[j] = (self.weights["alpha"][i, j].cpu() * channel) + self.weights["beta"][i, j].cpu()
            x[i] = hed
        x = self.run_mix(x)
        x = (x * self.max_input_value).float()
        return x

    def run_unmix(self, x):
        # operates in NCHW
        outputs = []
        for i, X in enumerate(x):
            Z = self._stain_unmix(X.permute(1, 2, 0), i)
            outputs.append(Z.permute(2, 0, 1))
        x = torch.stack(outputs)
        return x

    def run_mix(self, x):
        outputs = []
        for i, X in enumerate(x):
            Z = self._stain_remix(X.permute(1, 2, 0), i)
            outputs.append(Z.permute(2, 0, 1))
        x = torch.stack(outputs)
        return x

    def _stain_unmix(self, X, index):
        # operates in HWC
        X = torch.max(X, 1e-6 * torch.ones_like(X))
        Z = (torch.log(X) / np.log(1e-6)) @ torch.inverse(self.normalise_matrix(self.ruifrock))
        return Z

    def _stain_remix(self, Z, index):
        # operates in HWC
        log_adjust = -np.log(1E-6)
        log_rgb = -(Z * log_adjust) @ self.normalise_matrix(self.ruifrock)
        rgb = torch.exp(log_rgb)
        rgb = torch.clamp(rgb, 0, 1)
        return rgb

    def normalise_matrix(self, matrix):
        return fcn.normalize(matrix).float()


class RandomStainTransform(Transform):
    def __init__(self, input_shape, device, input_range=(0, 255)):
        super().__init__(input_shape, device)
        self.max_input_value = input_range[1]

        batch_size = self.input_shape[0]
        self.ruiford = np.array([[0.65, 0.70, 0.29],
                                 [0.07, 0.99, 0.11],
                                 [0.27, 0.57, 0.78]], dtype=np.float)
        self.base_weights = np.stack([self.ruiford] * batch_size)
        self.base_weights = torch.tensor(self.base_weights).to(device)

        self.weights = {
            "weights": torch.stack([torch.zeros((3, 3))] * batch_size).to(device),
        }

    def forward(self, x):
        x = self.run_unmix(x)
        x = self.run_mix(x)
        return x

    def run_unmix(self, x):
        # operates in NCHW
        x = (x / self.max_input_value).float()
        outputs = []
        for i, X in enumerate(x):
            Z = self._stain_unmix(X.permute(1, 2, 0), i)
            outputs.append(Z.permute(2, 0, 1))
        x = torch.stack(outputs)
        x = (x * self.max_input_value).float()
        return x

    def run_mix(self, x):
        x = (x / self.max_input_value).float()
        outputs = []
        for i, X in enumerate(x):
            Z = self._stain_remix(X.permute(1, 2, 0), i)
            outputs.append(Z.permute(2, 0, 1))
        x = torch.stack(outputs)
        x = (x * self.max_input_value).float()
        return x

    def _stain_unmix(self, X, index):
        # operates in HWC
        X = torch.max(X, 1e-6 * torch.ones_like(X))
        Z = (torch.log(X) / np.log(1e-6)) @ torch.inverse(self.normalise_matrix(self.base_weights[index]))
        return Z

    def _stain_remix(self, Z, index):
        # operates in HWC
        log_adjust = -np.log(1E-6)
        log_rgb = -(Z * log_adjust) @ self.normalise_matrix(self.base_weights[index] + self.weights["weights"][index])
        rgb = torch.exp(log_rgb)
        rgb = torch.clamp(rgb, 0, 1)
        return rgb

    def normalise_matrix(self, matrix):
        return fcn.normalize(matrix).float()


class JPEGTransform(Transform):
    def __init__(self, input_shape, device):
        super().__init__(input_shape, device)
        self.rounding = diff_round
        batch_size = self.input_shape[0]
        self.weights = {
            "quality": torch.stack([torch.tensor([100.0])] * batch_size).to("cpu")
        }

    def forward(self, x):
        '''
        '''
        x = x.to("cpu")
        for key in self.weights:
            self.weights[key] = self.weights[key].to("cpu")

        for i, _ in enumerate(x):
            self.factor = quality_to_factor(self.weights["quality"][i].item())
            self.compress = compress_jpeg(rounding=self.rounding, factor=self.factor)
            self.decompress = decompress_jpeg(self.input_shape[2], self.input_shape[3], rounding=self.rounding,
                                              factor=self.factor)
            x_i = x[i].unsqueeze(0) / 255
            y, cb, cr = self.compress(x_i)
            x_i = self.decompress(y, cb, cr)
            x[i] = x_i[0] * 255

        x = x.to(self.device)
        for key in self.weights:
            self.weights[key] = self.weights[key].to(self.device)
        return x



import torch

class Constraints:
    def l0(self, weights, base_weights, proportion):
        perturbation = weights - base_weights
        gradients = abs(weights.grad)
        flat = torch.flatten(gradients)
        k = int(proportion * len(flat))
        top_k = torch.topk(flat, k)
        if len(top_k[0]) == 0:
            threshold = 256
        else:
            threshold = top_k[0][-1]
        with torch.no_grad():
            perturbation[gradients < threshold] = 0
            weights += -weights + base_weights + perturbation
        return weights

    def l1(self, weights, base_weights, max_mae):
        perturbation = weights - base_weights
        with torch.no_grad():
            mae = torch.mean(torch.abs(perturbation))
            if mae > max_mae:
                multiplier = max_mae / mae
                perturbation *= multiplier
            weights += -weights + base_weights + perturbation
        return weights

    def l2(self, weights, base_weights, max_rmse):
        perturbation = weights - base_weights
        with torch.no_grad():
            rmse = torch.sqrt(torch.mean(torch.pow(perturbation, 2)))
            if rmse > max_rmse:
                multiplier = max_rmse / rmse
                perturbation *= multiplier
            weights += -weights + base_weights + perturbation
        return weights

    def l_inf(self, weights, base_weights, max_diff):
        perturbation = weights - base_weights
        with torch.no_grad():
            perturbation[:] = torch.clamp(perturbation, -max_diff, max_diff)
            weights += -weights + base_weights + perturbation
        return weights

    def in_range(self, weights, base_weights, range):
        return torch.clamp(weights, *range)
